from uuid import uuid4
from azure.data.tables import TableServiceClient, TableClient, UpdateMode
from azure.core.exceptions import ResourceExistsError
from datetime import datetime, timezone
import aiohttp
import os
import logging


IS_DEV = os.getenv("IS_DEV") in ["True", "true", "1", "TRUE"]
IS_TEST = os.getenv("IS_TEST") in ["True", "true", "1", "TRUE"]

logger = logging.getLogger(__name__)

def get_table_service():
    from bmsdna.mailing.config import get_default_credential
    short_azurite_constr = "UseDevelopmentStorage=true"
    azurite_con_str = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"
    
    if constr := os.getenv("AZURE_TABLE_CONNECTION_STRING"):
        if constr == short_azurite_constr:
            constr = azurite_con_str
        return TableServiceClient.from_connection_string(constr)
    elif endpoint := os.getenv("AZURE_TABLE_ENDPOINT"):
        return TableServiceClient(endpoint=endpoint, credential=get_default_credential())
    elif account_name := os.getenv("MAIL_STORAGE_ACCOUNT"):
        return TableServiceClient(endpoint=f"https://{account_name}.table.core.windows.net", credential=get_default_credential())
    elif IS_DEV or IS_TEST:
        return TableServiceClient.from_connection_string(azurite_con_str)
    else:
        raise ValueError("No AZURE_TABLE_CONNECTION_STRING or AZURE_TABLE_ENDPOINT provided")

class PersistanceDB:
    def __init__(self, read_only: bool = False) -> None:
        self.read_only = read_only
        self.db_dirty = False

    def init_db(self):
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        try:
            self.period_mail_log_table.create_table()
        except ResourceExistsError:
            pass
        except Exception as err:
            logger.warning("Could not create table. Assuming it exists", exc_info=err)
            
        try:
            self.mail_sent_log_table.create_table()
        except ResourceExistsError:
            pass
        except Exception as err:
            logger.warning("Could not create table. Assuming it exists", exc_info=err)

    def commit(self):
        pass

    async def __aenter__(self, *args, **kwargs):
        self._table_service = get_table_service()
        
        self.period_mail_log_table = self._table_service.get_table_client("periodMailLog")
        self.mail_sent_log_table = self._table_service.get_table_client("mailSentLog")

        self.init_db()
        return self

    async def _create_parquets(self):
        import polars as pl
        from .storage import _get_container_client
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        async with _get_container_client() as contc:
            since = None
            async for blob in contc.list_blobs("mailparquets/period_mail_log"):
                since = blob["last_modified"] if since is None else max(since, blob["last_modified"])

            query_filter = f"start ge datetime'{since.isoformat()}'" if since else "start ge datetime'2000-01-01T00:00:00Z'"
            entities = self.period_mail_log_table.query_entities(query_filter)
            all_df = pl.DataFrame([entity for entity in entities])

            temp_file = os.path.join(os.getenv("TEMP", "/tmp"), str(uuid4()) + ".parquet")
            if not all_df.is_empty():
                all_df.write_parquet(temp_file)
                now = datetime.now(tz=timezone.utc).strftime("%Y/%m/%d/%H%M/%S_")
                async with contc.get_blob_client(f"mailparquets/period_mail_log/{now}.parquet") as bc:
                    with open(temp_file, "rb") as f:
                        await bc.upload_blob(f, overwrite=True)
                os.remove(temp_file)

            since = None
            async for blob in contc.list_blobs("mailparquets/mail_sent_log"):
                since = blob["last_modified"] if since is None else max(since, blob["last_modified"])

            query_filter = f"sent_at ge datetime'{since.isoformat()}'" if since else "sent_at ge datetime'2000-01-01T00:00:00Z'"
            entities = self.mail_sent_log_table.query_entities(query_filter)
            all_df = pl.DataFrame([entity for entity in entities])

            temp_file = os.path.join(os.getenv("TEMP", "/tmp"), str(uuid4()) + ".parquet")
            if not all_df.is_empty():
                all_df.write_parquet(temp_file)
                now = datetime.now(tz=timezone.utc).strftime("%Y/%m/%d/%H%M/%S_")
                async with contc.get_blob_client(f"mailparquets/mail_sent_log/{now}.parquet") as bc:
                    with open(temp_file, "rb") as f:
                        await bc.upload_blob(f, overwrite=True)
                os.remove(temp_file)

    async def commit_to_storage(self, with_parquets: bool = True, reopen=True):
        if self.db_dirty and with_parquets:
            await self._create_parquets()
            self.db_dirty = False

    def save_execution_start(self, run_id: str, system_name: str, periodic_mail_name: str):
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        if self.read_only:
            raise ValueError("Cannot execute in readonly")
        self.db_dirty = True
        entity = {
            "PartitionKey": system_name + "_" + periodic_mail_name,
            "RowKey": run_id,
            "start": datetime.now(tz=timezone.utc)
        }
        self.period_mail_log_table.create_entity(entity)

    def get_logs(self, since: datetime):
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        query_filter = f"start ge datetime'{since.isoformat()}'"
        entities = self.period_mail_log_table.query_entities(query_filter)
        return [entity for entity in entities]

    def save_execution_end(self, run_id: str, system_name: str, periodic_mail_name: str, nr_failed: int, nr_ok: int, nr_skipped: int):
        self.db_dirty = True
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        if self.read_only:
            raise ValueError("Cannot execute in readonly")
        self.period_mail_log_table.update_entity({
            "PartitionKey": system_name + "_" + periodic_mail_name,
            "RowKey": run_id,
                "end": datetime.now(tz=timezone.utc),
                "nr_failed": nr_failed,
                "nr_ok": nr_ok,
                "nr_skipped": nr_skipped

        })

    def get_last_execution_start(self, system_name: str, periodic_mail_name: str, tzone: timezone):
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        part_key = system_name + "_" + periodic_mail_name
        query_filter = f"PartitionKey eq '{part_key}'"
        entities = self.period_mail_log_table.query_entities(query_filter, select=["start"])
        last_start = None
        for entity in entities:
            start = datetime.fromisoformat(entity["start"]).astimezone(tzone) if not isinstance(entity["start"], datetime) else entity["start"].astimezone(tzone)
            if not last_start or start > last_start:
                last_start = start
        return last_start

    async def was_mail_sent(self, session: aiohttp.ClientSession, system_name: str, periodic_mail_name: str, entity: str, entity_id: str, after: datetime):
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        part_key =system_name + "_" + periodic_mail_name
        row_key = entity + "_" + entity_id
        query_filter = f"ok eq true and PartitionKey eq '{part_key}' and RowKey eq '{row_key}'"
        entities = self.mail_sent_log_table.query_entities(query_filter)
        if any(entities):
            return "azure_table"

        from bmsdna.mailing.sendgrid import get_is_mail_sent
        try:
            if await get_is_mail_sent(session, system_name, periodic_mail_name, entity_id):
                return "sendgrid"
        except aiohttp.ClientResponseError:
            logger.warning("Error while checking mail sent status")
            return False
        return False

    def save_mail_sent(self, run_id: str, system_name: str, periodic_mail_name: str, entity: str, entity_id: str, ok: bool, status: int):
        self.db_dirty = True
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None
        if self.read_only:
            raise ValueError("Cannot execute in readonly")
        self.mail_sent_log_table.create_entity({
            "PartitionKey": system_name + "_" + periodic_mail_name,
            "RowKey": entity + "_" + entity_id,
            "run_id": run_id,
            "ok": ok,
            "status": status,
            "sent_at": datetime.now(tz=timezone.utc)
        })

    def get_last_execution_end(self, system_name: str, periodic_mail_name: str, tzone: timezone):
        part_key =system_name + "_" + periodic_mail_name
        assert self.period_mail_log_table is not None
        assert self.mail_sent_log_table is not None

        query_filter = f"PartitionKey eq '{part_key}'"
        entities = self.period_mail_log_table.query_entities(query_filter, select=["end"])
        last_end = None 
        for entity in entities:
            if end_dts := entity.get("end"):
                if end_dts is None:
                    continue
                end_dt = datetime.fromisoformat(end_dts).astimezone(tzone) if not isinstance(end_dts, datetime) else end_dts.astimezone(tzone)
                if not last_end or end_dt > last_end:
                    last_end = end_dt
        return last_end

    async def __aexit__(self, *args, **kwargs):
        await self.commit_to_storage(reopen=False)

        if self.period_mail_log_table is not None:
            self.period_mail_log_table.close()
        if self.mail_sent_log_table is not None:
            self.mail_sent_log_table.close()
        if self._table_service is not None:
            self._table_service.close()
        self.period_mail_log_table = None
        self.mail_sent_log_table = None
        self._table_service = None




from bmsdna.mailing.config import DEFAULT_SENDER
from .message import BaseAttachment, FileAttachment, MailMessage, get_loaded_attachment_or_error
from email import generator, message_from_binary_file
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
import mimetypes
import base64
from io import IOBase
from typing import TypeVar, cast
import logging

logger = logging.getLogger(__name__)

# see also https://www.rfc-editor.org/rfc/rfc2076


async def _create_eml_email(m: MailMessage):
    msg = MIMEMultipart()
    to_str = m.to if isinstance(m.to, str) else (", ".join(m.to) if m.to else None)

    if (not to_str or len(to_str) < 4) and (m.cc is not None or m.bcc is not None):
        to_str = DEFAULT_SENDER # azure does require a To Recipient
        assert DEFAULT_SENDER is not None
    real_to = to_str or DEFAULT_SENDER
    real_sender = m.sender or DEFAULT_SENDER
    assert real_to, "Must set a recipient or configure DEFAULT_SENDER"
    assert real_sender, "Must set a sender or configure DEFAULT_SENDER"
    msg["To"] = real_to
    msg["From"] = real_sender
    if m.cc is not None:
        msg["cc"] = m.cc if isinstance(m.cc, str) else ", ".join(m.cc)
    if m.bcc is not None:
        msg["bcc"] = m.bcc if isinstance(m.bcc, str) else ", ".join(m.bcc)
    msg["Subject"] = m.subject
    msg.attach(MIMEText(m.body, "html" if m.body_is_html else "plain"))

    if m.attachments is not None:
        for att in m.attachments:
            real_att, _ = await get_loaded_attachment_or_error(att)

            if not isinstance(real_att, FileAttachment):
                raise ValueError("Can only support file attachments in EML")
            assert isinstance(real_att.content, bytes)

            content_type, encoding = (real_att.content_type, None) or mimetypes.guess_type(real_att.filename)
            if content_type is None:
                content_type = "application/octet-stream"
            content_type_main, content_type_sub = content_type.split("/")
            if content_type_main == "image":  # pdf, excel etc
                msg_att = MIMEImage(real_att.content, content_type_sub)
            elif content_type_main == "text":  # pdf, excel etc
                msg_att = MIMEText(real_att.content.decode("utf-8"), content_type_sub)
            elif content_type_main == "application":  # pdf, excel etc
                msg_att = MIMEApplication(real_att.content, content_type_sub)
            else:  # we don't know this. octet-stream is generic
                msg_att = MIMEApplication(real_att.content)

            if real_att.content_id:
                msg_att.add_header("Content-Disposition", "inline")
                msg_att.add_header("Content-Id", real_att.content_id)
            else:
                msg_att.add_header("Content-Disposition", f'attachment; filename="{real_att.filename}"')
            msg.attach(msg_att)
    return msg


async def mail_to_base64eml(m: MailMessage):
    msg = await _create_eml_email(m)
    return base64.b64encode(msg.as_bytes())


T = TypeVar("T")


def _not_none(v: T | None) -> T:
    assert v is not None
    return v


def mail_from_eml(eml: bytes, entity: str, entity_id: str, system: str):
    import io
    from email.message import Message

    with io.BytesIO(eml) as iop:
        msg = message_from_binary_file(iop)

        payloads: list[Message] = msg.get_payload()  # type: ignore

        body_str: str = ""
        body_is_html = False
        atts: list[BaseAttachment] = []

        for item in payloads:
            if item.is_multipart():
                raise ValueError("Don't know multi part")
            content_dist = item.get_content_disposition()
            if content_dist is None:
                binary = item.get_payload(decode=True)  # type: ignore
                cnttype = item.get_content_type()
                charset = item.get_content_charset("utf-8")
                body_str = binary.decode(charset)
                body_is_html = "html" in cnttype
            elif content_dist == "attachment":
                binary: bytes = item.get_payload(decode=True)  # type: ignore
                filename = item.get_filename() or "file"
                from .attachments.static_attachment import StaticAttachment

                atts.append(StaticAttachment(binary, filename))
            elif content_dist == "inline":
                binary: bytes = item.get_payload(decode=True)  # type: ignore
                content_id = _not_none(item.get("Content-Id"))
                from .attachments.static_attachment import StaticInlineAttachment

                atts.append(StaticInlineAttachment(content=binary, content_id=content_id))
            else:
                raise ValueError(f"Cannot handle this: {content_dist}")
        sender = msg.get("From", DEFAULT_SENDER)
        assert sender is not None, "No sender found, configure DEFAULT_SENDER"
        return MailMessage(
            entity=entity,
            entity_id=entity_id,
            system=system,
            sender=sender,
            subject=_not_none(msg.get("Subject")),
            body=body_str,
            body_is_html=body_is_html,
            attachments=atts,
            bcc=[cast(str, i).strip() for i in msg.get("bcc", "").split(",") if len(i.strip()) > 0],
            cc=[cast(str, i).strip() for i in msg.get("cc", "").split(",") if len(i.strip()) > 0],
            to=_not_none(msg.get("To")),
        )


def _split_mails(mails: str) -> list[str]:
    return [m.strip() for m in mails.replace(";", ",").split(",") if m.strip()]


async def mail_to_jsondict_sendgrid(m: MailMessage):
    # see https://www.twilio.com/docs/sendgrid/api-reference/mail-send/mail-send
    tos = _split_mails(m.to) if isinstance(m.to, str) else m.to
    tos = [t.strip() for t in tos if t and "@" in t and t.strip()] if tos else []
    if len(tos) == 0 and (m.cc is not None or m.bcc is not None):
        assert DEFAULT_SENDER is not None, "No sender found, configure DEFAULT_SENDER"
        tos = [DEFAULT_SENDER]  # azure does require a To Recipient
    d = {
        "personalizations": [
            {
                "from": {
                    "email": m.sender,
                    "name": m.sender_name or m.sender,
                },
                "to": [{"email": item.strip()} for item in tos if item.strip()],
            }
        ],
        "from": {
            "email": m.sender,
            "name": m.sender_name or m.sender,
        },
        "subject": m.subject,
        "content": [{"type": "text/html" if m.body_is_html else "text/plain", "value": m.body}],
    }
    has_additional_recipients = False
    d["custom_args"] = {"system": m.system, "entity": m.entity, "entity_id": m.entity_id}
    if m.cc is not None:
        ccs = _split_mails(m.cc) if isinstance(m.cc, str) else m.cc
        d["personalizations"][0]["cc"] = [{"email": item.strip()} for item in ccs if item.strip() and item.strip() not in tos]
        has_additional_recipients = len(d["personalizations"][0]["cc"]) > 0
    if m.bcc is not None:
        bccs = _split_mails(m.bcc) if isinstance(m.bcc, str) else m.bcc
        d["personalizations"][0]["bcc"] = [{"email": item.strip()} for item in bccs if item.strip() and item.strip() not in tos]
        has_additional_recipients = has_additional_recipients or len(d["personalizations"][0]["bcc"]) > 0

    if len(d["personalizations"][0]["to"]) == 0 and has_additional_recipients:
        assert DEFAULT_SENDER is not None, "No sender found, configure DEFAULT_SENDER"
        d["personalizations"][0]["to"] = [{"email": DEFAULT_SENDER}]

    if m.attachments and len(m.attachments) > 0:
        atts = []
        for att in m.attachments:
            r_att, _ = await get_loaded_attachment_or_error(att)
            atts.append(r_att.serialize4json("sendgrid"))
        d["attachments"] = atts
    return d


async def mail_to_jsondict_graph(m: MailMessage):
    tos = _split_mails(m.to) if isinstance(m.to, str) else m.to
    tos = [t for t in tos if t and "@" in t] if tos else []
    if len(tos) == 0 and (m.cc is not None or m.bcc is not None):
        assert DEFAULT_SENDER is not None, "No sender found, configure DEFAULT_SENDER"
        tos = [DEFAULT_SENDER]  # azure does require a To Recipient
    d = {
        "subject": m.subject,
        "body": {"contentType": "Html" if m.body_is_html else "Text", "content": m.body},
        "toRecipients": [{"emailAddress": {"address": item.strip()}} for item in tos if item.strip()],
        "internetMessageHeaders": [
            {"name": "X-Entity", "value": m.entity},
            {"name": "X-Entity-Id", "value": m.entity_id},
            {"name": "X-System", "value": m.system},
        ],
    }
    if m.cc is not None:
        ccs = _split_mails(m.cc) if isinstance(m.cc, str) else m.cc
        d["ccRecipients"] = [{"emailAddress": {"address": item.strip()}} for item in ccs if item.strip()]
    if m.bcc is not None:
        bccs = _split_mails(m.bcc) if isinstance(m.bcc, str) else m.bcc
        d["bccRecipients"] = [{"emailAddress": {"address": item.strip()}} for item in bccs if item.strip()]

    if len(d["toRecipients"]) == 0 and (len(d.get("ccRecipients", [])) > 0 or len(d.get("bccRecipients", [])) > 0):
        assert DEFAULT_SENDER is not None, "No sender found, configure DEFAULT_SENDER"
        d["toRecipients"] = [{"emailAddress": {"address": DEFAULT_SENDER}}]

    if m.importance is not None:
        d["importance"] = m.importance.lower()
    if m.attachments and len(m.attachments) > 0:
        atts = []
        for att in m.attachments:
            r_att, _ = await get_loaded_attachment_or_error(att)
            atts.append(r_att.serialize4json("msgraph"))
        d["attachments"] = atts
    return {"message": d, "saveToSentItems": True}


async def write_eml_file(msg: MailMessage, out: IOBase):
    m = await _create_eml_email(msg)
    emlGenerator = generator.BytesGenerator(out)
    emlGenerator.flatten(m)


def replace_cid(msg: MailMessage, body: str):
    if 'src="cid:' in body:
        assert msg.attachments is not None
        ind = body.index('src="cid:')
        endind = body.index('"', ind + 9)
        cid = body[ind + 9 : endind]
        at = next(
            (a for a in msg.attachments if a.is_inline() and isinstance(a, FileAttachment) and a.content_id == cid)
        )
        fn, type_str, bytes = at.serialize4bytes()
        dataurl = "data:" + type_str + ";base64," + base64.b64encode(bytes).decode("ascii")
        body = body.replace('src="cid:' + cid + '"', 'src="' + dataurl + '"')
        return replace_cid(msg, body)

    return body


def get_debug_html(msg: MailMessage):
    from html import escape

    def render_line(title: str, value: str | list[str] | None):
        if value is None:
            return ""
        if isinstance(value, list):
            return "<p>" + escape(title) + ": " + ", ".join((escape(i) for i in value)) + "</p>"
        return "<p>" + escape(title) + ": " + escape(value) + "</p>"

    def render_mail(msg: MailMessage):
        html = ""
        html += "<h3>" + escape(msg.subject) + "</h3>"
        html += render_line("From", msg.sender)
        html += render_line("To", msg.to)
        html += render_line("CC", msg.cc)
        html += render_line("Bcc", msg.bcc)
        html += render_line("System", msg.system)
        html += render_line("entity", msg.entity)
        html += render_line("entity_id", msg.entity_id)
        html += render_line("Importance", msg.importance)
        if msg.attachments and any((a for a in msg.attachments if not a.is_inline())):
            html += "<br/><p>Attachments:</p><div>"
            for at in msg.attachments:
                if not at.is_inline():
                    filename, content_type, content = at.serialize4bytes()
                    data_url = f"data:{content_type};base64," + base64.b64encode(content).decode("ascii")
                    html += f'<a href="{data_url}">{escape(filename or "attachment")}</a>'
            html += "</div><br />"
        if msg.body_is_html:
            html += (
                '<br /> <br /><iframe style="min-height: 500px;width: 100%;min-width: 500px;border: 0" border=0 srcdoc="'
                + escape(replace_cid(msg, msg.body))
                + '"></iframe>'
            )
        else:
            html += "<br /> <br /><p>" + escape(msg.body) + "</p>"
        return html

    html = "<html>"
    html += "<head>"
    html += "<title>" + escape(msg.subject) + "</title>"
    html += """<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.1/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-4bw+/aepP/YC94hEpVNVgiZdgIC5+VKNBQNGCHeKRQN+PtmoHDEXuppvnDJzQIu9" crossorigin="anonymous">
<script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.1/dist/js/bootstrap.bundle.min.js" integrity="sha384-HwwvtgBNo3bZJJLYd8oVXjrBZt8cqVSpeBNS5n7C8IVInixGAoxmnlMuBnhbgrkm" crossorigin="anonymous"></script>"""
    html += "</head>"
    html += '<body style="padding: 10px">'
    html += render_mail(msg)
    html += "</body>"
    return html



import os
from typing import cast, TYPE_CHECKING

import pytest
from dotenv import load_dotenv

load_dotenv()

if TYPE_CHECKING:
    from docker.models.containers import Container


def start_azurite() -> "Container":
    import docker
    import docker.errors
    from docker.models.containers import Container

    client = docker.from_env()  # code taken from https://github.com/fsspec/adlfs/blob/main/adlfs/tests/conftest.py#L72
    azurite_server: Container | None = None
    try:
        m = cast(Container, client.containers.get("test4azurite"))
        if m.status == "running":
            return m
        else:
            azurite_server = m
    except docker.errors.NotFound:
        pass

    if azurite_server is None:
        azurite_server = client.containers.run(
            "mcr.microsoft.com/azure-storage/azurite:latest",
            detach=True,
            name="test4azurite",
            ports={"10000/tcp": 10000, "10001/tcp": 10001, "10002/tcp": 10002},
        )  # type: ignore
    assert azurite_server is not None
    azurite_server.start()
    print(azurite_server.status)
    print("Successfully created azurite container...")
    return azurite_server


@pytest.fixture(scope="session", autouse=True)
def spawn_azurite():
    if os.getenv("NO_AZURITE_DOCKER", "0") == "1":
        yield None
    else:
        azurite = start_azurite()
        yield azurite
        if os.getenv("KEEP_AZURITE_DOCKER", "0") == "0":  # can be handy during development
            azurite.stop()

@pytest.fixture(scope="session", autouse=True)
def init_azurite(spawn_azurite):
    from azure.data.tables import TableServiceClient
    azurite_con_str = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"
    table_service_client = TableServiceClient.from_connection_string(azurite_con_str)
    try:
        table_service_client.get_table_client("periodMailLog").delete_table()
    except Exception:
        pass
    try:
        table_service_client.get_table_client("mailSentLog").delete_table()
    except Exception:
        pass


import aiohttp
from .persistance import PersistanceDB



class GenerationContext:
    _http_sessions: dict[str, aiohttp.ClientSession] = dict()

    db: PersistanceDB

    def __init__(self, read_only: bool = False):
        self.db = PersistanceDB(read_only=read_only)
        self.caption_dict = None

    async def __aenter__(self, *args, **kwargs):
        await self.db.__aenter__(args, kwargs)
        return self

    def get_http_session(self, base_url: str | None = None):
        key = base_url.lower() if base_url else "__default__"
        res = self._http_sessions.get(key, None)
        if res:
            if not res.closed:
                return res
        res = aiohttp.ClientSession(base_url)
        self._http_sessions[key] = res
        return res


    async def __aexit__(self, *args, **kwargs):
        await self.db.__aexit__(*args, **kwargs)
        for _, v in self._http_sessions.items():
            await v.__aexit__(*args, **kwargs)
        self._http_sessions = {}



import aiohttp
import os
from azure.identity.aio import ManagedIdentityCredential, ClientSecretCredential
from azure.identity import InteractiveBrowserCredential
from typing import Union
from datetime import datetime, timezone

import logging

logger = logging.getLogger(__name__)

GRAPH_API_SCOPE = "https://graph.microsoft.com/.default"
POWERBI_API_SCOPE = "https://analysis.windows.net/powerbi/api/.default"


async def acquire_token(*, prefix="MAILER", scopes: Union[tuple[str], str, None] = None):
    """
    Acquire token via MSAL
    """
    prefix = prefix.upper()
    client_id = os.getenv(f"{prefix}_CLIENT_ID", None)
    client_secret = os.getenv(f"{prefix}_CLIENT_SECRET", None)
    cred: Union[ManagedIdentityCredential, ClientSecretCredential]
    if scopes is None:
        scopes = (GRAPH_API_SCOPE,)
    elif isinstance(scopes, str):
        scopes = (scopes,)
    assert scopes is not None

    if not client_id:
        if os.getenv("IS_DEV", "0") == "1":
            scred = InteractiveBrowserCredential()
            tk = scred.get_token(*scopes)
            return (tk.token, datetime.fromtimestamp(tk.expires_on, tz=timezone.utc))
        cred = ManagedIdentityCredential()
    elif not client_secret:
        cred = ManagedIdentityCredential(client_id=client_id)
    else:
        tenant_id = os.getenv("TENANT_ID", os.getenv("AZURE_TENANT_ID"))
        assert tenant_id, "AZURE_TENANT_ID must be set"
        cred = ClientSecretCredential(
            tenant_id,
            client_id=client_id,
            client_secret=client_secret,
        )
    async with cred:
        tk = await cred.get_token(*scopes)
        return (tk.token, datetime.fromtimestamp(tk.expires_on, tz=timezone.utc))


async def get_paged_request_data(session: aiohttp.ClientSession, headers: dict, starturl: str, pagesize=900):
    starturl_with_top = starturl + ("&" if "?" in starturl else "?") + "$top=" + str(pagesize)
    url = starturl_with_top  # f"https://api.powerbi.com/v1.0/myorg/groups/{grpid}/users?$top=900"  # paging does not seem to be supported very well
    skip = 0
    total_items = -1
    while url is not None:
        async with session.get(url, headers=headers) as user_dt:
            if not user_dt.ok:
                logger.error(f"Failure calling {url}: \r\n" + (await user_dt.text()))
                user_dt.raise_for_status()
            users: dict = await user_dt.json()
            dtlist = users["value"]
            nrusers = len(dtlist)
            yield dtlist
            url = None
            if "@odata.nextLink" in users:
                url = users["@odata.nextLink"]
            elif "@odata.count" in users:
                total_items = users["@odata.count"]
            if total_items > (skip + nrusers) and url is None:
                skip = skip + pagesize
                url = starturl_with_top + "&$skip=" + str(skip)



from functools import lru_cache
import os
from typing import Any, Optional
from azure.core.credentials import TokenCredential, AccessToken
from azure.core.credentials_async import AsyncTokenCredential

DEFAULT_SENDER = os.getenv("MAIL_DEFAULT_SENDER")
DEFAULT_SENDER_NAME:str | None = os.getenv("MAIL_DEFAULT_SENDER_NAME", DEFAULT_SENDER)

class TokenAdapter(AsyncTokenCredential):
    def __init__(self, sync_cred: TokenCredential):
        super().__init__()
        self.sync_cred = sync_cred

    
    async def get_token(
        self,
        *args,
        **kwargs: Any,
    ) -> AccessToken:
        """Request an access token for `scopes`.

        :param str scopes: The type of access needed.

        :keyword str claims: Additional claims required in the token, such as those returned in a resource
            provider's claims challenge following an authorization failure.
        :keyword str tenant_id: Optional tenant to include in the token request.
        :keyword bool enable_cae: Indicates whether to enable Continuous Access Evaluation (CAE) for the requested
            token. Defaults to False.

        :rtype: AccessToken
        :return: An AccessToken instance containing the token string and its expiration time in Unix time.
        """
        return self.sync_cred.get_token(*args, **kwargs)

    async def close(self) -> None:
        pass

    async def __aexit__(
        self,
        *args,
        **kwargs
    ) -> None:
        pass

def get_async_default_credential() -> AsyncTokenCredential:
    if cred := os.getenv("USE_DR_SERVICE_CREDENTIAL"):
        from databricks.sdk.runtime import dbutils
        
        return  TokenAdapter(dbutils.credentials.getServiceCredentialsProvider(cred)) # type: ignore
    else:
        
        from azure.identity.aio import DefaultAzureCredential
        return DefaultAzureCredential()

@lru_cache(1)
def get_default_credential():
    if cred := os.getenv("USE_DR_SERVICE_CREDENTIAL"):
        from databricks.sdk.runtime import dbutils
        return dbutils.credentials.getServiceCredentialsProvider(cred) # type: ignore
    else:
        
        from azure.identity import DefaultAzureCredential
        return DefaultAzureCredential()
    



from .message import MailMessage
from .message_io import write_eml_file
import os
from datetime import timezone, datetime
import uuid
import io
import aiofiles
import logging

logger = logging.getLogger(__name__)

container_name = "mail"


def _get_container_client():
    from azure.storage.blob.aio import ContainerClient
    from bmsdna.mailing.config import get_async_default_credential

    account_name = os.getenv("MAIL_STORAGE_ACCOUNT")
    if account_name:
        return ContainerClient(account_name + ".blob.core.windows.net", container_name, credential=get_async_default_credential())
    constr = os.getenv("MAIL_STORAGE_CONNECTION", os.getenv("AzureWebJobsStorage"))
    if constr == "UseDevelopmentStorage=true":
        constr = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;QueueEndpoint=http://127.0.0.1:10001/devstoreaccount1;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"
    assert constr is not None
    if os.getenv("IS_DEV") == "1":
        from azure.storage.blob import ContainerClient as ContainerClientSync

        with ContainerClientSync.from_connection_string(constr, container_name) as cs:
            if not cs.exists():
                cs.create_container()
    return ContainerClient.from_connection_string(constr, container_name)


def _get_container_client_sync():
    from azure.storage.blob import ContainerClient as ContainerClientSync
    from bmsdna.mailing.config import get_default_credential


    if account_name := os.getenv("MAIL_STORAGE_ACCOUNT"):
        return ContainerClientSync(account_name + ".blob.core.windows.net", container_name, credential=get_default_credential())
    constr = os.getenv("MAIL_STORAGE_CONNECTION", os.getenv("AzureWebJobsStorage"))
    if constr == "UseDevelopmentStorage=true":
        constr = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;QueueEndpoint=http://127.0.0.1:10001/devstoreaccount1;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"
    assert constr is not None
    if os.getenv("IS_DEV") == "1":
        with ContainerClientSync.from_connection_string(constr, container_name) as cs:
            if not cs.exists():
                cs.create_container()
    return ContainerClientSync.from_connection_string(constr, container_name)


async def save_sent_mail(m: MailMessage):
    async with _get_container_client() as contc:
        datepath = datetime.now().strftime("%Y/%m/%d/%H%M")
        emailid = str(m.email_id or uuid.uuid4())
        path = f"sent/{m.system}/{m.entity}/{datepath}/{m.entity_id}_{emailid}.eml"
        async with contc.get_blob_client(path) as bc:
            with io.BytesIO() as bio:
                await write_eml_file(m, bio)
                bio.seek(0)
                await bc.upload_blob(bio)


async def save_mail_failure(
    m: MailMessage, status: int, binary_response: bytes, content_type: str, json_data: str | None
):
    async with _get_container_client() as contc:
        datepath = datetime.now().strftime("%Y/%m/%d/%H%M")
        emailid = str(m.email_id or uuid.uuid4())
        path = f"error/{m.system}/{m.entity}/{datepath}/{m.entity_id}_{emailid}.eml"
        path_error = f"error/{m.system}/{m.entity}/{datepath}/{m.entity_id}_{emailid}.txt"
        path_error2 = f"error/{m.system}/{m.entity}/{datepath}/{m.entity_id}_{emailid}.json"
        async with contc.get_blob_client(path) as bc:
            with io.BytesIO() as bio:
                await write_eml_file(m, bio)
                bio.seek(0)
                await bc.upload_blob(bio)
        async with contc.get_blob_client(path_error) as bc:
            await bc.upload_blob(binary_response)
        if json_data is not None:
            async with contc.get_blob_client(path_error2) as bc:
                await bc.upload_blob(json_data)

    logger.error("Could not send mail: " + binary_response.decode("utf-8"))


async def get_database(db_name: str, force=False, old=False):
    async with _get_container_client() as contc:
        path = db_name
        localpath = os.getenv("TEMP", "/tmp") + "/" + db_name
        os.makedirs(localpath, exist_ok=True)
        try:
            if not await contc.exists():
                await contc.create_container()
        except Exception:
            pass
        async for item in contc.list_blobs(path):
            assert item.name is not None
            if item.name.endswith(".old") and not old:
                continue
            if old and not item.name.endswith(".old"):
                continue
            assert item.last_modified is not None
            last_mod: datetime = item.last_modified
            if last_mod.tzinfo is None:
                last_mod = last_mod.astimezone(timezone.utc)

            fname = item.name.split("/")[-1]
            full_local_path = os.path.join(localpath, fname)
            if old and full_local_path.endswith(".old"):
                full_local_path = full_local_path[:-4]
            if os.path.exists(full_local_path) and not force:
                lmod = datetime.fromtimestamp(os.path.getmtime(full_local_path))
                if lmod.tzinfo is None:
                    lmod = lmod.astimezone(timezone.utc)
                if lmod > last_mod:
                    continue
            res = await contc.download_blob(item.name)
            async with aiofiles.open(full_local_path, "wb") as f:
                b = await res.read(4048)
                while b:
                    await f.write(b)
                    b = await res.read(4048)
        return localpath + "/" + db_name + ".sqlite"


def get_database_sync(db_name: str):
    with _get_container_client_sync() as contc:
        path = db_name
        localpath = os.getenv("TEMP", "/tmp") + "/" + db_name
        os.makedirs(localpath, exist_ok=True)
        for item in contc.list_blobs(path):
            assert item.name is not None
            assert item.last_modified is not None
            last_mod: datetime = item.last_modified
            if last_mod.tzinfo is None:
                last_mod = last_mod.astimezone(timezone.utc)

            fname = item.name.split("/")[-1]
            full_local_path = os.path.join(localpath, fname)
            if os.path.exists(full_local_path):
                lmod = datetime.fromtimestamp(os.path.getmtime(full_local_path))
                if lmod.tzinfo is None:
                    lmod = lmod.astimezone(timezone.utc)
                if lmod > last_mod:
                    continue
            res = contc.download_blob(item.name)
            with open(full_local_path, "wb") as f:
                b = res.read(4048)
                while b:
                    f.write(b)
                    b = res.read(4048)
        return localpath + "/" + db_name + ".sqlite"


async def write_database(db_name: str):
    async with _get_container_client() as contc:
        localpath = os.getenv("TEMP", "/tmp") + "/" + db_name
        os.makedirs(localpath, exist_ok=True)
        for item in os.listdir(localpath):
            if not item.endswith(".old"):
                full_path = os.path.join(localpath, item)
                async with aiofiles.open(full_path, "rb") as f:
                    async with contc.get_blob_client(db_name + "/" + item) as cl:
                        if await cl.exists():
                            async with contc.get_blob_client(db_name + "/" + item + ".old") as cl2:
                                await cl2.upload_blob(await (await cl.download_blob()).readall(), overwrite=True)

                        await cl.upload_blob(f, overwrite=True)



import asyncio
from datetime import datetime, timezone
import os
from typing import Awaitable, Callable, Literal, TypeVar
import aiohttp


async def retry_many(func: Callable[[], Awaitable[aiohttp.ClientResponse]]) ->aiohttp.ClientResponse:
    for i in range(9):
        rsp = await func()
        if rsp.status == 429:
            await asyncio.sleep(2 ** i)
        else:
            return rsp
    return await func()

async def get_is_mail_sent(http_session: aiohttp.ClientSession, system: str, mail: str, entity_id: str):
    api_key = os.getenv("SENDGRID_API_KEY")
    if not api_key:
        return False
    headers = {"Authorization": "Bearer " + api_key}
    assert '"' not in system
    assert '"' not in mail
    assert '"' not in entity_id
    query = f"(unique_args['system']=\"{system}\") AND (unique_args['entity']=\"{mail}\") AND (unique_args['entity_id']=\"{entity_id}\")"
    async def _load_base():
        return await http_session.get("https://api.sendgrid.com/v3/messages", params={"query": query, "limit": "2"}, headers=headers)
    rsp = await retry_many(_load_base) 
    rsp.raise_for_status()
    jsd = await rsp.json()
    return len(jsd["messages"]) > 0


async def get_last_sent_date(
    http_session: aiohttp.ClientSession,
    system: str,
    mail: str,
    since: datetime | str,
    stop_if_gte: datetime | None = None,
) -> datetime | None | Literal["invalid"]:
    api_key = os.getenv("SENDGRID_API_KEY")
    if not api_key:
        return "invalid"
    since_str = (
        since.astimezone(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ") if not isinstance(since, str) else since
    )
    query = f'(unique_args[\'system\']="{system}") ANd (unique_args[\'entity\']="{mail}") AND last_event_time >= TIMESTAMP "{since_str}"'
    headers = {"Authorization": "Bearer " + api_key}
    async def _load_base():
        return await http_session.get(
            "https://api.sendgrid.com/v3/messages", params={"query": query, "limit": 1000}, headers=headers
        )
    rsp = await retry_many(_load_base)
    if rsp.status >= 300:
        print(await rsp.text())
    rsp.raise_for_status()
    jsd = await rsp.json()
    if len(jsd["messages"]) == 0:
        return None
    max_date = None
    msg_cnt = 0
    all_msgs : list[dict]= jsd["messages"]
    all_msgs.sort(key=lambda x: datetime.fromisoformat(x["last_event_time"]), reverse=True)
    for msg in all_msgs:
        if msg["status"] == "processing":
            date_str = msg["last_event_time"]
        else:
            msg_cnt += 1
            if msg_cnt > 10:
                return max_date
            rsp_msg = await http_session.get(f"https://api.sendgrid.com/v3/messages/{msg['msg_id']}", headers=headers)
            if rsp_msg.status == 429:  # I'm not getting a price for this, but hey, it works
                await asyncio.sleep(10)
                rsp_msg = await http_session.get(
                    f"https://api.sendgrid.com/v3/messages/{msg['msg_id']}", headers=headers
                )
                if rsp_msg.status == 404: # no longer there, a bit strange, but ok
                    continue
                if rsp_msg.status == 429:
                    await asyncio.sleep(20)
                    rsp_msg = await http_session.get(
                        f"https://api.sendgrid.com/v3/messages/{msg['msg_id']}", headers=headers
                    )
                    if rsp_msg.status == 429:
                        await asyncio.sleep(60)
                        rsp_msg = await http_session.get(
                            f"https://api.sendgrid.com/v3/messages/{msg['msg_id']}", headers=headers
                        )
                        if rsp_msg.status == 404:
                            continue    
                        if rsp_msg.status == 429:
                            return max_date

            rsp_msg.raise_for_status()
            msg_js = await rsp_msg.json()
            process_date = next((x for x in msg_js["events"] if x["event_name"] == "processed"))["processed"]
            date_str = process_date
        date_time = datetime.fromisoformat(date_str)
        if max_date is None or date_time > max_date:
            max_date = date_time
            if stop_if_gte is not None and max_date > stop_if_gte:
                return max_date
    return max_date



from pathlib import Path
from ..message import FileAttachment


class StaticAttachment(FileAttachment):
    def __init__(self, content: bytes | Path, filename: str) -> None:
        super().__init__()
        if isinstance(content, Path):
            content = content.read_bytes()
        self.content = content
        self._filename = filename

    def get_content(self):
        assert self.content is not None
        return self.content


class StaticInlineAttachment(FileAttachment):
    def __init__(self, *, content: bytes | Path, content_id: str) -> None:
        super().__init__()
        if isinstance(content, Path):
            self.content = content.read_bytes()
        else:
            assert isinstance(content, bytes)
            self.content = content
        self.content_id = content_id
        self._filename = "inline_" + content_id

    async def get_content(self) -> bytes:
        assert self.content is not None
        return self.content



import base64
import mimetypes
import os
from typing import Any, Literal, Optional, Union, cast, Awaitable, TYPE_CHECKING
from dataclasses import dataclass
import dataclasses
import inspect
from abc import ABC, abstractmethod
import logging

import aiohttp

from bmsdna.mailing.config import DEFAULT_SENDER, DEFAULT_SENDER_NAME

if TYPE_CHECKING:
    from .sending import SendResult

logger = logging.getLogger(__name__)


class BaseAttachment(ABC):
    @property
    @abstractmethod
    def filename(self) -> str: ...

    @abstractmethod
    def get_error_message(self) -> str: ...

    @abstractmethod
    async def assure_persisted_content(self) -> int: ...

    @abstractmethod
    def serialize4json(self, flavor: Literal["msgraph", "sendgrid"]) -> dict: ...

    @abstractmethod
    def serialize4bytes(self) -> tuple[str | None, str, bytes]: ...

    @abstractmethod
    def is_inline(self) -> bool: ...


def _remove_none(d: dict) -> dict:
    return {k: v for k, v in d.items() if v is not None}


class FileAttachment(BaseAttachment):
    _filename: str
    # content: Union[bytes, Callable[["FileAttachment"], bytes], Callable[["FileAttachment"], Awaitable[bytes]]]
    content: bytes | None = None
    content_id: Optional[str] = None  # use for in-mail references
    content_type: Optional[str] = None

    @property
    def filename(self):
        return self._filename

    def get_error_message(self) -> str:
        return f"Error loading attachment {self.filename}"

    def is_inline(self) -> bool:
        return self.content_id is not None

    @abstractmethod
    def get_content(self) -> bytes | Awaitable[bytes]: ...

    def serialize4json(self, flavor: Literal["msgraph", "sendgrid"]) -> dict:
        content_type = self.content_type or mimetypes.guess_type(self.filename)[0]
        assert isinstance(self.content, bytes)
        b64_content = base64.b64encode(self.content).decode("ascii")
        if flavor == "msgraph":
            return _remove_none(
                {
                    "@odata.type": "#microsoft.graph.fileAttachment",
                    "name": self.filename,
                    "isInline": True if self.content_id else None,
                    "contentId": self.content_id,
                    "contentType": content_type,
                    "contentBytes": b64_content,
                }
            )
        elif flavor == "sendgrid":
            return _remove_none(
                {
                    "filename": self.filename,
                    "content": b64_content,
                    "type": content_type,
                    "disposition": "attachment" if not self.content_id else "inline",
                    "content_id": self.content_id,
                }
            )
        else:
            raise ValueError(f"Unsupported flavor {flavor}")

    def serialize4bytes(self) -> tuple[str | None, str, bytes]:
        content_type = self.content_type or mimetypes.guess_type(self.filename)[0]
        assert isinstance(self.content, bytes)
        return (self.filename, content_type or "application/octet-stream", self.content)

    async def assure_persisted_content(self) -> int:
        if isinstance(self.content, bytes):
            return len(self.content)
        real_content = self.get_content()
        if inspect.isawaitable(real_content):
            real_content2 = cast(bytes, await real_content)
        else:
            real_content2 = cast(bytes, real_content)
        assert isinstance(real_content2, bytes)
        self.content = real_content2
        return len(self.content)


async def get_loaded_attachment_or_error(att: BaseAttachment) -> tuple[BaseAttachment, bool]:
    real_att = att
    has_error = False
    try:
        await att.assure_persisted_content()
    except Exception as err:
        from bmsdna.mailing.attachments.static_attachment import StaticAttachment

        msg = att.get_error_message()
        real_att = StaticAttachment(msg.encode("utf-8"), f"error_{att.filename}.txt")
        logger.error(f"Error loading attachment {att.filename}: {err}")
        has_error = True
    return real_att, has_error


# see also https://www.rfc-editor.org/rfc/rfc2076
@dataclass(frozen=True)
class MailMessage:
    entity: str
    entity_id: str
    system: str

    subject: str
    body: str
    body_is_html: bool
    to: Union[list[str], str]
    email_id: str | int | None = None
    sender: str | None = DEFAULT_SENDER
    sender_name: str | None = DEFAULT_SENDER_NAME
    cc: Union[list[str], str, None] = None
    bcc: Union[list[str], str, None] = None
    attachments: Union[list[BaseAttachment], None] = None
    importance: Optional[Literal["High", "normal", "low"]] = "normal"

    extra_infos: dict[str, Any] | None = None

    send_method: Optional[Literal["SENDGRID"]] = None

    error_receiver: Optional[str] = None

    async def assure_attachments_loaded(self) -> "MailMessage":
        if not self.attachments:
            return self
        new_atts: list[BaseAttachment] = []
        has_error = False
        for att in self.attachments:
            att2, has_error2 = await get_loaded_attachment_or_error(att)
            new_atts.append(att2)
            has_error = has_error or has_error2
        if has_error and self.error_receiver is not None:
            return dataclasses.replace(self, attachments=new_atts, to=self.error_receiver, bcc=None, cc=None)

        return dataclasses.replace(self, attachments=new_atts)

    async def send(self, session: aiohttp.ClientSession) -> "SendResult":
        from .sending import send_mail

        return await send_mail(self, session)


def mail_message_from_json(json_str: str):
    import json

    d = json.loads(json_str)
    return MailMessage(**d)



from typing import Literal
from ..message import BaseAttachment


# THIS IS UNTESTED AND MS-GRAPH ONLY
class ReferenceAttachment(BaseAttachment):
    def __init__(
        self,
        name: str,
        *,
        source_url: str,
        provider_type: Literal["oneDriveBusiness", "oneDriveConsumer", "dropbox", "other"],
        permission: Literal[
            "other", "view", "edit", "anonymousView", "anonymousEdit", "organizationView", "organizationEdit"
        ],
        is_folder: bool = False,
    ) -> None:
        super().__init__()
        self.name = name
        self.source_url = source_url
        self.provider_type = provider_type
        self.permission = permission
        self.is_folder = is_folder

    def get_error_message(self) -> str:
        # not really needed
        return f"Error loading attachment {self.name} from {self.source_url}"

    async def assure_persisted_content(self):
        return 1  # nothing to persist here

    def serialize4json(self, flavor: Literal["msgraph", "sendgrid"]) -> dict:
        if flavor == "sendgrid":
            raise ValueError("ReferenceAttachment is not supported in SendGrid")
        return {
            "@odata.type": "#microsoft.graph.referenceAttachment",
            "name": self.name,
            "sourceUrl": self.source_url,
            "providerType": self.provider_type,
            "permission": self.permission,
            "isFolder": self.is_folder,
        }



from typing import NamedTuple, Literal

from bmsdna.mailing.config import DEFAULT_SENDER
from .basicazure import acquire_token
from .message import MailMessage
from .message_io import mail_to_jsondict_graph, mail_to_jsondict_sendgrid
from .storage import save_sent_mail, save_mail_failure
import aiohttp
import json
import os
import logging

logger = logging.getLogger(__name__)

_env_var_send_method = os.getenv("MAILING_SEND_METHOD", "SENDGRID").upper()
assert _env_var_send_method in [
    "SENDGRID",
], f"Invalid value for MAILING_SEND_METHOD: {_env_var_send_method}"
SEND_METHOD: Literal["SENDGRID"] = _env_var_send_method  # type: ignore


class SendResult(NamedTuple):  # noqa: F821
    ok: bool
    status: int
    content: str | bytes
    content_type: str


async def send_mail(mail: MailMessage, http_session: aiohttp.ClientSession) -> SendResult:
    mail = await mail.assure_attachments_loaded()
    method = send_mail_using_sendgrid  # for now
    return await method(mail, http_session)


async def send_mail_using_sendgrid(m: MailMessage, http_session: aiohttp.ClientSession) -> SendResult:
    if os.getenv("IS_DEV", "0") == "1" or os.getenv("IS_TEST", "0") == "1":
        return SendResult(True, 200, "fake".encode("utf-8"), "text/plain")
    assert m.subject is not None
    assert m.body is not None
    headers = {
        "Authorization": "Bearer " + os.environ["SENDGRID_API_KEY"],
        "Content-Type": "application/json",
    }
    jsd = await mail_to_jsondict_sendgrid(m)
    rsp = await http_session.post("https://api.sendgrid.com/v3/mail/send", data=json.dumps(jsd), headers=headers)
    bcontent = await rsp.content.read()
    if rsp.ok:
        logger.info(f"Mail sent: {m.subject}", extra={"system": m.system, "entity": m.entity, "entity_id": m.entity_id, "to": m.to, "cc": m.cc, "bcc": m.bcc})
        await save_sent_mail(m)
    else:
        logger.warning(f"Mail sent: {m.subject}", extra={"system": m.system,
                                                         "status": rsp.status,
                                                         "content": bcontent.decode("utf-8"),
                                                          "entity": m.entity, "entity_id": m.entity_id, "to": m.to, "cc": m.cc, "bcc": m.bcc},
                       )
        await save_mail_failure(m, rsp.status, bcontent, rsp.content_type, json_data=json.dumps(jsd))
    return SendResult(rsp.ok, rsp.status, bcontent, rsp.content_type)


async def send_mail_using_js(m: MailMessage, http_session: aiohttp.ClientSession) -> SendResult:
    if os.getenv("IS_DEV", "0") == "1" or os.getenv("IS_TEST", "0") == "1":
        return SendResult(True, 200, "fake".encode("utf-8"), "text/plain")
    assert m.subject is not None
    assert m.body is not None
    base_url = f"https://graph.microsoft.com/v1.0/users/{m.sender or DEFAULT_SENDER}/"

    token, expiry = await acquire_token()
    headers = {
        "Authorization": "Bearer " + token,
        "Content-Type": "application/json",
    }  # see https://learn.microsoft.com/en-us/graph/api/user-sendmail?view=graph-rest-1.0&tabs=http#example-4-send-a-new-message-using-mime-format

    jsd = await mail_to_jsondict_graph(m)
    rsp = await http_session.post(base_url + "sendMail", data=json.dumps(jsd), headers=headers)
    bcontent = await rsp.content.read()
    if rsp.ok:
        await save_sent_mail(m)
    else:
        await save_mail_failure(m, rsp.status, bcontent, rsp.content_type, json_data=json.dumps(jsd))
    return SendResult(rsp.ok, rsp.status, bcontent, rsp.content_type)



from dataclasses import dataclass
from uuid import uuid4

from bmsdna.mailing.config import DEFAULT_SENDER, DEFAULT_SENDER_NAME
from .scheduling.schedule import Weekday, Schedule, DeltaSchedule, INVALID_DAY_ACTIONS
import pytz
from datetime import date, datetime, timezone, time
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Final,
    Callable,
    Iterable,
    Literal,
    NamedTuple,
    Optional,
    TypedDict,
    Union,
    cast,
)
from datetime import timedelta
from .generation_context import GenerationContext
from dateutil.relativedelta import relativedelta
import holidays
import inspect
import time as time_module
from .message import BaseAttachment, MailMessage
import logging 

logger = logging.getLogger(__name__)

TZONE: Final[timezone] = pytz.timezone("Europe/Zurich")  # type: ignore
HOLIDAYS = holidays.country_holidays("CH")
year = datetime.now().year
for year in range(year, year + 3):  # holidays from 26 on
    HOLIDAYS.append(date(year, 12, 26))
    HOLIDAYS.append(date(year, 12, 27))
    HOLIDAYS.append(date(year, 12, 28))
    HOLIDAYS.append(date(year, 12, 29))
    HOLIDAYS.append(date(year, 12, 30))
    HOLIDAYS.append(date(year, 12, 31))


@dataclass(frozen=True)
class GenerationInfo:
    system: str
    mail_name: str
    dry_run: bool
    context: GenerationContext

    def make_mail(
        self,
        *,
        subject: str,
        entity_id: str,  # a unique id for the mail
        body: str,
        body_is_html: bool,
        to: Union[list[str], str],
        sender: str | None = DEFAULT_SENDER,
        sender_name: str | None = DEFAULT_SENDER_NAME,
        cc: Union[list[str], str, None] = None,
        bcc: Union[list[str], str, None] = None,
        attachments: "Union[list[BaseAttachment], None]" = None,
        importance: Optional[Literal["High", "normal", "low"]] = "normal",
        extra_infos: dict[str, Any] | None = None,
        send_method: Optional[Literal["SENDGRID"]] = None,
        error_receiver: Optional[str] = None,
    ) -> MailMessage:
        return MailMessage(
            entity=self.mail_name,
            entity_id=entity_id,
            system=self.system,
            subject=subject,
            body=body,
            body_is_html=body_is_html,
            to=to,
            sender=sender,
            sender_name=sender_name,
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            importance=importance,
            extra_infos=extra_infos,
            send_method=send_method,
            error_receiver=error_receiver,
        )


MailGenerationFunc = Callable[
    [GenerationInfo],
    MailMessage
    | Iterable[MailMessage]
    | None
    | Awaitable[MailMessage | Iterable[MailMessage] | None]
    | AsyncIterable[MailMessage],
]


@dataclass(frozen=True)
class RegistryInfo:
    func: MailGenerationFunc
    schedule_str: str
    retry_on_empty_send: bool
    schedule: Schedule | None


_registry: dict[tuple[str, str], RegistryInfo] = {}


_todos: list[tuple[str, str]] = []


def _get_send_mail_for_scheduler_fn(func: MailGenerationFunc, inf: tuple[str, str]):
    def todos():
        global _todos
        _todos.append(inf)

    return todos


def daily_mail(system: str, name: str, day_time: time, start_date: datetime, *, retry_on_empty_send: bool):
    def daily_decorator(func: MailGenerationFunc):
        _registry[(system, name)] = RegistryInfo(
            func,
            "daily",
            retry_on_empty_send=retry_on_empty_send,
            schedule=DailyTimerNoWe(TZONE, start_date, day_time),
        )
        return func

    return daily_decorator


def monthly_mail(
    system: str,
    name: str,
    start_date: datetime,
    day_time: time | None = None,
    every_n_months=1,
    *,
    on_invalid_day_action: INVALID_DAY_ACTIONS = "forward",
    retry_on_empty_send: bool,
):
    def monthly_decorator(func: MailGenerationFunc):
        _registry[(system, name)] = RegistryInfo(
            func,
            "monthly",
            retry_on_empty_send=retry_on_empty_send,
            schedule=DeltaSchedule(
                TZONE,
                start_date,
                None,
                day_time,
                relativedelta(months=every_n_months),
                on_invalid_day_action=on_invalid_day_action,
            ),
        )
        return func

    return monthly_decorator


def weekly_mail(
    system: str,
    name: str,
    day: Weekday,
    start_date: datetime,
    day_time: time | None = None,
    *,
    every_n_weeks=1,
    retry_on_empty_send: bool,
):
    if start_date.weekday() != day.value:
        raise ValueError("Start date must be the same day as the day of the week")

    def weekly_decorator(func: MailGenerationFunc):
        _registry[(system, name)] = RegistryInfo(
            func,
            "weekly",
            retry_on_empty_send=retry_on_empty_send,
            schedule=DeltaSchedule(TZONE, start_date, day, day_time, relativedelta(weeks=every_n_weeks)),
        )
        return func

    return weekly_decorator


class MonthlyTimerNoWe(DeltaSchedule):
    def __init__(
        self,
        tz: timezone,
        start: datetime,
        day_time: time | None,
        td: timedelta | relativedelta,
        day_of_month: int,
        on_invalid_day_action: INVALID_DAY_ACTIONS = "forward",
    ) -> None:
        super().__init__(
            tz, start, weekday=None, day_time=day_time, td=td, on_invalid_day_action=on_invalid_day_action
        )
        self.day_of_month = day_of_month

    def _next_date(self, dt: datetime) -> datetime:
        s = super()._next_date(dt)
        s2 = s.replace(day=self.day_of_month)
        return s2

    def is_day_allowed(self, day: datetime) -> bool:
        if day.weekday() >= 5:
            return False
        if day in HOLIDAYS:
            return False
        return super().is_day_allowed(day)


class DailyTimerNoWe(DeltaSchedule):
    def __init__(self, tz: timezone, start: datetime, day_time: time | None) -> None:
        super().__init__(
            tz,
            start,
            weekday=None,
            day_time=day_time,
            td=timedelta(days=1),
            on_invalid_day_action="forward",
        )

    def is_day_allowed(self, day: datetime) -> bool:
        if day.weekday() >= 5:
            return False
        if day in HOLIDAYS:
            return False
        return super().is_day_allowed(day)


def monthly_mail_no_we(
    system: str,
    name: str,
    start_date: datetime,
    day_of_month: int,
    day_time: time | None = None,
    *,
    every_n_months=1,
    on_invalid_day_action: INVALID_DAY_ACTIONS = "forward",
    retry_on_empty_send: bool,
):
    def monthly_decorator(func: MailGenerationFunc):
        _registry[(system, name)] = RegistryInfo(
            func,
            "monthly_no_we",
            retry_on_empty_send=retry_on_empty_send,
            schedule=MonthlyTimerNoWe(
                TZONE,
                start_date,
                day_of_month=day_of_month,
                day_time=day_time,
                td=relativedelta(months=every_n_months),
                on_invalid_day_action=on_invalid_day_action,
            ),
        )
        return func

    return monthly_decorator


def external_schedule_mail(system: str, name: str):
    def external_decorator(func: MailGenerationFunc):
        _registry[(system, name)] = RegistryInfo(func, "external", retry_on_empty_send=False, schedule=None)
        return func

    return external_decorator


class MailInfos(TypedDict):
    system: str
    mail_name: str
    schedule: Schedule | None


def get_mail_infos() -> list[MailInfos]:
    return [MailInfos(system=k[0], mail_name=k[1], schedule=v.schedule) for k, v in _registry.items()]


def get_function_by_alias(alias: str):
    for key, value in _registry.items():
        if key[0] + "_" + key[1] == alias:
            return (key[0], key[1])
    raise KeyError(f"Key {alias} not found")


class EmptyMailException(Exception):
    pass


class SendInfo(NamedTuple):
    nr_ok: int
    nr_failed: int
    nr_skipped: int


async def _mail_messages(
    res: MailMessage
    | Iterable[MailMessage]
    | None
    | Awaitable[MailMessage | Iterable[MailMessage] | None]
    | AsyncIterable[MailMessage]
    | None,
):
    if inspect.isasyncgen(res):
        async for mail in res:
            yield cast(MailMessage, mail)
        return
    real_res = await res if res is not None and isinstance(res, Awaitable) else res
    if real_res is None:
        return
    real_res = [real_res] if isinstance(real_res, MailMessage) else real_res or []
    for m in real_res:  # type: ignore
        yield cast(MailMessage, m)


async def _send_mails(
    context: GenerationContext, inf: GenerationInfo, reginf: RegistryInfo, *, max_runtime_in_seconds: float | None
):
    run_id = str(uuid4())
    start = time_module.time()
    context.db.save_execution_start(run_id, inf.system, inf.mail_name)
    res = reginf.func(inf)
    real_res = await res if res is not None and isinstance(res, Awaitable) else res
    if real_res is None:
        real_res = []
    real_res = [real_res] if isinstance(real_res, MailMessage) else real_res or []

    nr_failed = 0
    nr_ok = 0
    nr_skipped = 0
    nr_days_back = 1
    if (
        isinstance(reginf.schedule, DeltaSchedule)
        and isinstance(reginf.schedule.timedelta, relativedelta)
        and reginf.schedule.timedelta.weeks >= 1
    ):
        nr_days_back = 2
    yesterday_utc = datetime.now(TZONE).astimezone(timezone.utc) - timedelta(days=nr_days_back)
    last_saved: datetime | None = None
    had_mail = False
    async for mail in _mail_messages(real_res):
        had_mail = True
        assert inf.system == mail.system
        assert inf.mail_name == mail.entity
        if sent_source := await context.db.was_mail_sent(
            context.get_http_session(), inf.system, inf.mail_name, mail.entity, mail.entity_id, yesterday_utc
        ):
            logger.warning(f"Skipping {inf.system} {inf.mail_name} {mail.entity_id} as it was already sent ({sent_source})")
            nr_skipped += 1
            continue
        res = await mail.send(context.get_http_session())
        context.db.save_mail_sent(run_id, mail.system, inf.mail_name, mail.entity, mail.entity_id, res.ok, res.status)
        if res.ok:
            nr_ok += 1
        else:
            nr_failed += 1
        now = datetime.now(TZONE)
        if last_saved is None or (now - last_saved).total_seconds() > 180:
            await context.db.commit_to_storage(with_parquets=False)
            last_saved = now
        else:
            context.db.commit()

        if max_runtime_in_seconds is not None and time_module.time() - start > max_runtime_in_seconds:
            break

    if reginf.retry_on_empty_send and not had_mail:
        raise EmptyMailException()
    context.db.save_execution_end(
        run_id, inf.system, inf.mail_name, nr_failed=nr_failed, nr_ok=nr_ok, nr_skipped=nr_skipped
    )
    await context.db.commit_to_storage(with_parquets=False)
    context.db.commit()
    return SendInfo(nr_ok, nr_failed, nr_skipped)


async def get_mails_dry(
    context: GenerationContext, system: str, mail_name: str, *, filter: Callable[[MailMessage], bool] | None = None
):
    mail = _registry[(system, mail_name)]
    res = mail.func(GenerationInfo(system, mail_name, dry_run=True, context=context))
    real_res = _mail_messages(res)
    async for m in real_res:
        if filter is None or filter(m):
            yield await m.assure_attachments_loaded()


async def send_mails(context: GenerationContext, system: str, mail_name: str, *, max_runtime_in_seconds: float | None):
    mail = _registry[(system, mail_name)]
    return await _send_mails(
        context,
        GenerationInfo(system, mail_name, dry_run=False, context=context),
        mail,
        max_runtime_in_seconds=max_runtime_in_seconds,
    )


async def get_jobs_todo(dry: bool, context: GenerationContext, *, filter_system: str | None = None, filter_mail: str | None = None):
    from bmsdna.mailing.sendgrid import get_last_sent_date

    now = datetime.now(TZONE)
    for key, item in _registry.items():
        if item.schedule:
            if filter_system is not None and key[0] != filter_system:
                continue
            if filter_mail is not None and key[1] != filter_mail:
                continue
            before_date = item.schedule.calc_last_date(now)
            check_since = item.schedule.calc_last_date(before_date)
            ldt = context.db.get_last_execution_end(key[0], key[1], TZONE)
            sdt = context.db.get_last_execution_start(key[0], key[1], TZONE)
            if sdt and ldt and sdt > ldt:
                ldt = sdt  # if the last execution was not finished, we assume it is still running and will finish
            elif not ldt and sdt:
                ldt = sdt
            if item.schedule.is_due(ldt): # it's due, we double-check if it was sent by sendgrid
                send_grid_date = await get_last_sent_date(
                context.get_http_session(), key[0], key[1], check_since, before_date + timedelta(days=1)
                )
                if send_grid_date == "invalid":
                    send_grid_date = None
                if send_grid_date is not None and (ldt is None or send_grid_date > ldt):
                    ldt = send_grid_date
            if item.schedule.is_due(ldt):
                yield GenerationInfo(key[0], key[1], dry_run=dry, context=context)


def get_next_executions(context: GenerationContext):
    res = []
    for key, item in _registry.items():
        ldt = context.db.get_last_execution_end(key[0], key[1], TZONE)
        if item.schedule:
            exec_date = item.schedule.calc_next_date(ldt)
        else:
            exec_date = None
        res.append(
            {
                "system": key[0],
                "mail_name": key[1],
                "last_executed": ldt,
                "exec_date": exec_date,
            }
        )
    return res



from email import message_from_file
from email.message import Message

with open("out/file.eml") as f:
    msg = message_from_file(f)

for name, t in msg.items():
    print(name)
    print(t)

payloads: list[Message] = msg.get_payload()  # type: ignore
for item in payloads:
    if item.is_multipart():
        raise ValueError("Don't know that one")
    content_dist = item.get_content_disposition()
    if content_dist is None:
        binary = item.get_payload(decode=True)  # type: ignore
        cnttype = item.get_content_type()
        charset = item.get_content_charset("utf-8")
        strvl = binary.decode(charset)
        print(strvl)
    elif content_dist == "attachment":
        binary: bytes = item.get_payload(decode=True)[:10]  # type: ignore
        print(binary)
    elif content_dist == "inline":
        # well, how to handle that?
        print(item.get_payload(decode=True)[:10])  # type: ignore



from datetime import datetime, time, timezone, timedelta
from abc import ABC, abstractmethod
from typing import Literal
from enum import Enum
from dateutil.relativedelta import relativedelta

INVALID_DAY_ACTIONS = Literal["forward", "backward"]


class Weekday(Enum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6


WeeksDays_No_WE = [Weekday.MONDAY, Weekday.TUESDAY, Weekday.WEDNESDAY, Weekday.THURSDAY, Weekday.FRIDAY]


class Schedule(ABC):
    def __init__(self, tz: timezone, start: datetime) -> None:
        super().__init__()
        self.tz = tz
        self.start = start.astimezone(tz) if start.tzinfo is None else start

    def is_due(self, last_exec: datetime | None, *, now: datetime | None = None) -> bool:
        now = now or datetime.now(tz=self.tz)
        if last_exec is not None and (now - last_exec).total_seconds() < 60 * 15:
            return False  # last execution was less than 15 minutes ago, that is not possible
        return self.calc_next_date(last_exec=last_exec, now=now) <= now

    @abstractmethod
    def calc_next_date(self, last_exec: datetime | None, *, now: datetime | None = None) -> datetime:
        pass

    @abstractmethod
    def calc_last_date(self, from_date: datetime) -> datetime:
        pass


class DeltaSchedule(Schedule):
    def __init__(
        self,
        tz: timezone,
        start: datetime,
        weekday: Weekday | list[Weekday] | None,
        day_time: time | None,
        td: timedelta | relativedelta,
        on_invalid_day_action: INVALID_DAY_ACTIONS = "forward",
    ) -> None:
        super().__init__(tz, start)
        self.weekdays = weekday
        self.day_time = (
            day_time
            if not day_time or day_time.tzinfo
            else time(hour=day_time.hour, minute=day_time.minute, second=day_time.second, tzinfo=tz)
        )
        self.weekdays_int = (
            [c.value for c in (weekday if isinstance(weekday, list) else [weekday])] if weekday else None
        )
        self.timedelta = td
        self.on_invalid_day_action = on_invalid_day_action

    def is_day_allowed(self, day: datetime) -> bool:
        if self.weekdays_int is None:
            return True
        return day.weekday() in self.weekdays_int

    def _real_start(self) -> datetime:
        real_start = self.start
        while not self.is_day_allowed(real_start):
            real_start = real_start + timedelta(days=1 if self.on_invalid_day_action == "forward" else -1)
        self.real_start = real_start
        return real_start

    def _next_date(self, dt: datetime) -> datetime:
        if self.day_time is not None:
            return (
                dt.replace(
                    hour=self.day_time.hour,
                    minute=self.day_time.minute,
                    second=self.day_time.second,
                    microsecond=self.day_time.microsecond,
                )
                + self.timedelta
            )
        return dt + self.timedelta

    def calc_last_date(self, from_date):
        if from_date < self.start:
            return self.start
        dt = from_date - self.timedelta
        while not self.is_day_allowed(dt):
            dt = dt - timedelta(days=-1)
        return dt

    def calc_next_date(self, last_exec: datetime | None, *, now: datetime | None = None) -> datetime:
        now = now or datetime.now(tz=self.tz)
        if now <= self._real_start() or last_exec is None:
            return self._real_start()
        if (now - last_exec).total_seconds() < 60 * 15:
            return last_exec + self.timedelta
        exec_date = self.start
        while exec_date < last_exec:
            exec_date = self._next_date(exec_date)
        if exec_date == last_exec:
            next_date = self._next_date(exec_date)
        else:
            next_date = exec_date
        next_date_orig = next_date
        while not self.is_day_allowed(next_date):
            next_date = next_date + timedelta(days=1 if self.on_invalid_day_action == "forward" else -1)
        if next_date <= last_exec:
            next_date = self._next_date(next_date_orig)
            while not self.is_day_allowed(next_date):
                next_date = next_date + timedelta(days=1 if self.on_invalid_day_action == "forward" else -1)

        if self.day_time:
            next_date = next_date.replace(
                hour=self.day_time.hour, minute=self.day_time.minute, second=self.day_time.second
            )
        return next_date



import asyncio

from pydantic import BaseModel
from ..message import FileAttachment
from typing import Literal, Optional, TypeAlias, TypedDict
import json
import aiohttp

is_pydantic_v1 = not hasattr(BaseModel, "model_json_schema")

class ExportError(TypedDict):
    code: str
    message: str


class ExportStatus(TypedDict):
    id: str
    percentComplete: int
    status: Literal["Failed", "NotStarted", "Running", "Succeeded", "Undefined"]
    resourceLocation: str
    expirationTime: str
    error: Optional[ExportError]


class ReportGenerationError(Exception):
    def __init__(self, error: ExportError | ExportStatus) -> None:
        super().__init__((error["code"] + ": " + error["message"] if "code" in error else json.dumps(error)))
        self.api_error = error


POWERBI_FORMATS: TypeAlias = Literal[
    "ACCESSIBLEPDF", "DOCX", "IMAGE", "MHTML", "PDF", "PNG", "PPTX", "XML", "PNG", "CSV", "XLSX"
]


class ParameterValue(BaseModel):
    name: str
    value: str


class PaginatedReportConfig(BaseModel):
    formatSettings: dict = {}
    identities: list[dict] = []
    parameterValues: list[ParameterValue] = []

    def set_parameter_value(self, name: str, value: str):
        has_param = False
        for p in self.parameterValues:
            if p.name == name:
                p.value = value
                has_param = True
        if not has_param:
            self.parameterValues.append(ParameterValue(name=name, value=value))
        return self

    def set_parameters(self, args_as_dict: Optional[dict[str, str]] = None, **kwargs):
        if args_as_dict:
            for k, v in args_as_dict.items():
                self.set_parameter_value(k, v)
        if kwargs:
            for k, v in kwargs.items():
                self.set_parameter_value(k, v)
        return self


class PowerBIPaginatedReportAttachment(FileAttachment):
    paginatedReportConfiguration: PaginatedReportConfig
    format: POWERBI_FORMATS = "PDF"

    def get_error_message(self) -> str:
        import urllib.parse

        msg = "Error generating PowerBI report.\r\n"
        msg += "You can download the report manually from the PowerBI service here:\r\n"
        msg += f"https://app.powerbi.com/groups/{self.group_id}/rdlreports/{self.report_id}"
        if self.paginatedReportConfiguration.parameterValues:
            msg += "?"
            msg += "&".join(
                (
                    f"rp:{p.name}={urllib.parse.quote_plus(p.value)}"
                    for p in self.paginatedReportConfiguration.parameterValues
                )
            )
        return msg

    def __init__(
        self,
        report_id: str,
        group_id: str,
        filename: str,
        *,
        format: POWERBI_FORMATS = "PDF",
        paginatedReportConfiguration: Optional[PaginatedReportConfig] = None,
        report_args: Optional[dict[str, str]] = None,
        api_token: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.report_id = report_id
        self.group_id = group_id
        self.format = format
        self.paginatedReportConfiguration = paginatedReportConfiguration or PaginatedReportConfig()
        if report_args:
            self.paginatedReportConfiguration.set_parameters(report_args)
        self._token = api_token
        self._filename = filename

    async def get_content(self) -> bytes:
        if isinstance(self.content, bytes):
            return self.content
        if not self._token:
            from ..basicazure import acquire_token, POWERBI_API_SCOPE

            token, exp_dt = await acquire_token(prefix="POWERBI", scopes=POWERBI_API_SCOPE)
            self._token = token
        headers = {"Authorization": "Bearer " + self._token}

        async with aiohttp.ClientSession() as session:
            config = self.paginatedReportConfiguration.model_dump() if not is_pydantic_v1 else self.paginatedReportConfiguration.dict()
            
            res = await session.post(
                f"https://api.powerbi.com/v1.0/myorg/groups/{self.group_id}/reports/{self.report_id}/ExportTo",
                headers=headers,
                json={
                    "format": self.format,
                    "paginatedReportConfiguration":config,
                },
            )
            if res.status > 299:
                print(await res.text())
            res.raise_for_status()
            jsd: ExportStatus = await res.json()
            export_id = jsd["id"] if "id" in jsd else None
            print(json.dumps(jsd))
            res.raise_for_status()
            if jsd["status"] == "Failed":
                raise ValueError(jsd["status"])
            else:
                status = jsd["status"]
                cnt = 0
                while status not in ["Failed", "Succeeded"]:
                    cnt += 1
                    await asyncio.sleep(10)
                    status_url = f"https://api.powerbi.com/v1.0/myorg/groups/{self.group_id}/reports/{self.report_id}/exports/{export_id}"
                    res = await session.get(status_url, headers=headers)
                    res.raise_for_status()
                    jsd: ExportStatus = await res.json()
                    if cnt == 1 or cnt % 10 == 0:
                        print(json.dumps(jsd))
                    status = jsd["status"]
            if status == "Succeeded":
                res = await session.get(jsd["resourceLocation"], headers=headers)
                res.raise_for_status()
                self.content = await res.read()
                return self.content
            elif status == "Failed":
                assert jsd is not None
                raise ReportGenerationError(jsd.get("error", jsd) or jsd)
            else:
                raise ValueError("Status not found")



from datetime import datetime
from datetime import time
from bmsdna.mailing.generation_context import GenerationContext
from bmsdna.mailing.message import MailMessage
from bmsdna.mailing.registry import send_mails
import pytest

from bmsdna.mailing.registry import TZONE, GenerationInfo, daily_mail


@pytest.mark.asyncio
async def test_double_send():
    @daily_mail(
        "test_system",
        "test_periodic_mail",
        start_date=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE),
        day_time=time(8, 0, tzinfo=TZONE),
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo):
        return MailMessage(
            entity="test_periodic_mail",
            entity_id="test_entity_id",
            system="test_system",
            subject="test_subject",
            body="test_body",
            body_is_html=False,
            to="holzöpfl_zipflchape@bmsuisse.ch",
        )

    async with GenerationContext() as genc:
        genc.db.init_db()
        si = await send_mails(genc, "test_system", "test_periodic_mail", max_runtime_in_seconds=20)
        assert si.nr_ok == 1
        assert si.nr_failed == 0
        assert si.nr_skipped == 0
        si = await send_mails(genc, "test_system", "test_periodic_mail", max_runtime_in_seconds=20)
        assert si.nr_ok == 0
        assert si.nr_failed == 0
        assert si.nr_skipped == 1

    async with GenerationContext() as genc:
        si = await send_mails(genc, "test_system", "test_periodic_mail", max_runtime_in_seconds=20)
        assert si.nr_ok == 0
        assert si.nr_failed == 0
        assert si.nr_skipped == 1



from datetime import datetime, time
from bmsdna.mailing.registry import (
    GenerationInfo,
    monthly_mail,
    TZONE,
    _registry,
    weekly_mail,
    monthly_mail_no_we,
    daily_mail,
)
from bmsdna.mailing.scheduling import Weekday


def test_monthly():
    @monthly_mail(
        "monthly",
        "test",
        datetime(2023, 8, 7, 6, 0, tzinfo=TZONE),
        day_time=time(6, 0, tzinfo=TZONE),
        on_invalid_day_action="backward",
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo) -> None:
        pass

    s = _registry[("monthly", "test")].schedule
    assert s is not None
    assert not s.is_due(None, now=datetime(2023, 8, 7, 5, 59, tzinfo=TZONE))
    assert s.is_due(None, now=datetime(2023, 8, 7, 6, 1, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 8, 7, 6, tzinfo=TZONE), now=datetime(2023, 8, 7, 7, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 8, 7, 7, tzinfo=TZONE), now=datetime(2023, 9, 7, 6, 1, tzinfo=TZONE))


def test_2weekly():
    @weekly_mail(
        "2weekly",
        "test",
        start_date=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE),
        day=Weekday.SUNDAY,
        day_time=time(8, 0, tzinfo=TZONE),
        every_n_weeks=2,
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo) -> None:
        pass

    s = _registry[("2weekly", "test")].schedule
    assert s is not None
    assert not s.is_due(None, now=datetime(2022, 12, 7, 5, 59, tzinfo=TZONE))
    assert s.is_due(None, now=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE))
    assert s.is_due(None, now=datetime(2023, 1, 1, 8, 1, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, 1, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, 20, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 8, 8, 1, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 13, 8, 1, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 15, 8, 0, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 15, 8, 1, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 1, 1, 8, tzinfo=TZONE), now=datetime(2023, 1, 15, 8, 0, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 1, 15, 8, tzinfo=TZONE), now=datetime(2023, 1, 29, 8, 0, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 15, 8, tzinfo=TZONE), now=datetime(2023, 1, 28, 8, 0, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 1, 15, 8, tzinfo=TZONE), now=datetime(2023, 1, 28, 8, 1, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 1, 15, 8, tzinfo=TZONE), now=datetime(2023, 1, 29, 8, 1, tzinfo=TZONE))


def test_monthly_no_we():
    @monthly_mail_no_we(
        "monthlynw",
        "test",
        start_date=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE),
        day_of_month=1,
        day_time=time(8, 0, tzinfo=TZONE),
        on_invalid_day_action="backward",
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo) -> None:
        pass

    s = _registry[("monthlynw", "test")].schedule
    assert s is not None
    assert s.is_due(None, now=datetime(2022, 12, 30, 8, 0, tzinfo=TZONE))  # 1.1 is a holiday, so is 31.12 -> 30.12
    assert not s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE))
    assert not s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, 1, tzinfo=TZONE))
    assert not s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 1, 1, 8, 20, tzinfo=TZONE))

    assert not s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 1, 31, 8, 0, tzinfo=TZONE))
    assert s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 2, 1, 8, 0, tzinfo=TZONE))
    assert s.is_due(datetime(2022, 12, 30, 8, 0, tzinfo=TZONE), now=datetime(2023, 2, 1, 8, 20, tzinfo=TZONE))
    assert not s.is_due(datetime(2023, 2, 1, 8, 20, tzinfo=TZONE), now=datetime(2023, 2, 28, 8, 1, tzinfo=TZONE))
    assert s.is_due(datetime(2023, 2, 1, 8, 20, tzinfo=TZONE), now=datetime(2023, 3, 1, 8, 1, tzinfo=TZONE))


def test_daily():
    @daily_mail(
        "daily_mailer",
        "test",
        day_time=time(6, 0, tzinfo=TZONE),
        start_date=datetime(2024, 3, 26, 6, 0, tzinfo=TZONE),
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo) -> None:
        pass

    s = _registry[("daily_mailer", "test")].schedule
    assert s is not None
    assert s.is_due(None, now=datetime(2024, 3, 26, 6, 0, tzinfo=TZONE))
    assert not s.is_due(datetime(2024, 3, 26, 6, 0, tzinfo=TZONE), now=datetime(2024, 3, 26, 7, 0, tzinfo=TZONE))
    assert not s.is_due(datetime(2024, 3, 26, 7, 0, tzinfo=TZONE), now=datetime(2024, 3, 26, 10, 0, tzinfo=TZONE))
    assert s.is_due(datetime(2024, 3, 26, 7, 0, tzinfo=TZONE), now=datetime(2024, 3, 27, 10, 0, tzinfo=TZONE))
    assert s.calc_next_date(datetime(2024, 3, 26, 7, 12, tzinfo=TZONE)) == datetime(2024, 3, 27, 6, 0, tzinfo=TZONE)



import time
from .storage import _get_container_client
from .registry import get_jobs_todo, send_mails
from .generation_context import GenerationContext
import logging
import sys
import traceback

logger = logging.getLogger(__name__)


async def run_jobs(
    max_runtime_in_seconds: int | None = None, *, filter_system: str | None = None, filter_mail: str | None = None
):
    from .registry import EmptyMailException

    errors = []
    all_oks = 0
    all_failes = 0
    all_empty = 0
    all_skipped = 0
    results = []
    start = time.time()
    async with _get_container_client() as contc:

        async with GenerationContext() as context:
            async for td in get_jobs_todo(dry=False, context=context, filter_system=filter_system, filter_mail=filter_mail):
                if filter_system is not None and td.system != filter_system:
                    continue
                if filter_mail is not None and td.mail_name != filter_mail:
                    continue
                async with contc.get_blob_client(f"lock_{td.system}__{td.mail_name}.txt") as bc:
                    if not await bc.exists():
                        await bc.upload_blob("empty")
                    async with await bc.acquire_lease(lease_duration=60) as lease:
                        try:
                            timeout_in_s = (
                                max_runtime_in_seconds - (time.time() - start)
                                if max_runtime_in_seconds is not None
                                else None
                            )
                            nr_ok, nr_failed, nr_skipped = await send_mails(
                                context, td.system, td.mail_name, max_runtime_in_seconds=timeout_in_s
                            )
                            results.append(
                                {
                                    "mail": f"{td.system}/{td.mail_name}",
                                    "ok": nr_ok,
                                    "failed": nr_failed,
                                    "skipped": nr_skipped,
                                }
                            )
                            all_oks += nr_ok
                            all_failes += nr_failed
                            all_skipped += nr_skipped
                        except EmptyMailException:
                            logger.warning(f"No mails to send for {td.system}/{td.mail_name}")
                            results.append(
                                {"mail": f"{td.system}/{td.mail_name}", "ok": 0, "failed": 0, "empty": "empty"}
                            )
                            all_empty += 1
                        except Exception as err:
                            logger.error("Error in send_mails", exc_info=err)
                            errors.append(err)
                            results.append(
                                {"mail": f"{td.system}/{td.mail_name}", "ok": 0, "failed": 1, "err": str(err)}
                            )
                            async with contc.get_blob_client(
                                f"last_run_error/{td.system}/{td.mail_name}_error.txt"
                            ) as bc:
                                err_str = repr(err) + "\r\n\r\n" + traceback.format_exc()
                                await bc.upload_blob(err_str, overwrite=True)
                        if max_runtime_in_seconds is not None and time.time() - start > max_runtime_in_seconds:
                            logger.warning(f"Max runtime of {max_runtime_in_seconds} seconds reached")
                            await context.db.commit_to_storage(with_parquets=True)
                            break
                        else:
                            await lease.renew()
                            await context.db.commit_to_storage(with_parquets=False)
                    await context.db.commit_to_storage(with_parquets=True)

    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        if sys.version_info > (3, 11):
            raise ExceptionGroup("Errors in run_jobs", errors)
        else:
            raise errors[0]
    elif all_failes > 0:
        raise Exception(f"{all_failes} mails failed to send. See Logs in funcstore/error")
    return results



from datetime import datetime, timezone
from datetime import time
from typing import AsyncIterable, TypeVar
from bmsdna.mailing.generation_context import GenerationContext
from bmsdna.mailing.persistance import PersistanceDB
from bmsdna.mailing.registry import get_jobs_todo
import pytest
import os

from bmsdna.mailing.registry import TZONE, GenerationInfo, monthly_mail_no_we
from bmsdna.mailing.storage import _get_container_client, get_database

T = TypeVar("T")


async def _list(l: AsyncIterable[T]):
    return [x async for x in l]


@pytest.mark.asyncio
async def test_db():
    @monthly_mail_no_we(
        "system_name",
        "periodic_mail_name",
        start_date=datetime(2023, 1, 1, 8, 0, tzinfo=TZONE),
        day_of_month=1,
        day_time=time(8, 0, tzinfo=TZONE),
        on_invalid_day_action="backward",
        retry_on_empty_send=False,
    )
    def _generate_mail(info: GenerationInfo) -> None:
        pass

    os.environ["MAIL_STORAGE_CONNECTION"] = "UseDevelopmentStorage=true"
    async with _get_container_client() as cc:
        try:
            await cc.delete_container()
        except:
            pass
        await cc.create_container()
    lcpath = await get_database("maildb")
    try:
        os.unlink(lcpath)
    except:
        pass
    async with GenerationContext() as genc:
        pers_db = genc.db

        todos = await _list(get_jobs_todo(True, genc))
        tm = [t for t in todos if t.system == "system_name" and t.mail_name == "periodic_mail_name"]
        assert len(tm) == 1

        start_db = datetime.now(tz=timezone.utc)
        pers_db.save_execution_start("run_id", "system_name", "periodic_mail_name")

        l1 = pers_db.get_logs(start_db)
        assert len(l1) == 1
        assert l1[0]["RowKey"] == "run_id"
        assert l1[0]["PartitionKey"] == "system_name_periodic_mail_name"
        assert l1[0].get("nr_failed") is None
        assert l1[0].get("nr_ok") is None
        assert l1[0].get("start") is not None
        assert l1[0].get("end") is  None

        todos = await _list(get_jobs_todo(True, genc))
        tm = [t for t in todos if t.system == "system_name" and t.mail_name == "periodic_mail_name"]
        assert len(tm) == 0

        pers_db.save_execution_end("run_id", "system_name", "periodic_mail_name", 1, 3, nr_skipped=5)
        l = pers_db.get_logs(start_db)
        assert len(l) == 1
        assert l[0]["RowKey"] == "run_id"
        assert l[0]["PartitionKey"] == "system_name_periodic_mail_name"
        assert l[0]["nr_failed"] == 1
        assert l[0]["nr_ok"] == 3
        assert l[0]["nr_skipped"] == 5
        assert l[0]["start"] is not None
        assert l[0]["end"] is not None
        last_end =  l[0]["end"]

        todos = await _list(get_jobs_todo(True, genc))
        tm = [t for t in todos if t.system == "system_name" and t.mail_name == "periodic_mail_name"]
        
        assert len(tm) == 0
        le = pers_db.get_last_execution_end("system_name", "periodic_mail_name", TZONE)
        assert le ==  last_end

    try:
        os.unlink(lcpath)
    except:
        pass
    async with (
        GenerationContext() as genc
    ):  # should still be zero, as we do need to refresh the file from blob storage
        todos = await _list(get_jobs_todo(True, genc))
        tm = [t for t in todos if t.system == "system_name" and t.mail_name == "periodic_mail_name"]
        assert len(tm) == 0
